class X7EEG(object):
    def __init__(self, data) :
        self._data = data
        self.info = {}

    def _parse(self, data):        
        self.info["packet_id"] = int.from_bytes(data[:8], 'little')
        self.info["channels"] = data[17:49]
        self.info["origin_rate"] = int.from_bytes(data[49:53], 'little')
        self.info["sample_rate"] = int.from_bytes(data[53:57], 'little')
        self.info["sample_length"] = int.from_bytes(data[57:61], 'little')
        self.info["resolution"] = int.from_bytes(data[61:63], 'little')
        self.info["filter"] = int.from_bytes(data[63], 'little')
        self.info["date_len"] = int.from_bytes(data[64:68], 'little')

class EEGContainer(object):
    def __init__(self) -> None:
        pass

    def append(self, data):
        eeg = X7EEG(data)
        return





from ctypes.wintypes import HICON
import struct

def bytes_to_float(ba):
    return struct.unpack('f', ba)[0]

def bytes_to_number(bs, isBig=True):
    return int.from_bytes(bs, byteorder='little', signed=isBig)

def to_long(bs, isBig=True):
    return int.from_bytes(bs, byteorder='little', signed=isBig)

class Message(object):
    def __init__(self) -> None:
        pass

    def parse(packet) -> None:
        pass






from queue import Queue
from threading import Thread
from loguru import logger

from qldev.utils import crc16, get_ip


class UDPMessage(object):
    MessageStart = 'SHQuanLan'
    ConnectCommand = 0x10
    DeviceInfoCommand = 0x09

    def __init__(self) -> None:
        pass

    @staticmethod
    def get_connect(devno, tcp_port=19128):
        message = bytearray(28)
        message[:10] = UDPMessage.MessageStart.encode('utf-8')
        cmd = UDPMessage.ConnectCommand
        devtype = 0x0
        
        # 本机ip
        ip = get_ip().split(".")
        
        message[10:12] = cmd.to_bytes(2, 'little')
        message[12:14] = devtype.to_bytes(2, 'little')
        #d57ee4dc99ce0940        
        message[14:22] = bytes.fromhex(devno)
        message[22] = (int)(ip[0])
        message[23] = (int)(ip[1])
        message[24] = (int)(ip[2])
        message[25] = (int)(ip[3])
        message[26:28] = tcp_port.to_bytes(2, 'little')
        checksum = crc16(message)

        return message + checksum.to_bytes(2, 'little')


class MessageChannel(Queue):
    def __init__(self, maxsize: int = ...) -> None:
        super().__init__(maxsize)
        self._run = True


    def subscribe(self, callback):
        consumer_thread = Thread(target=self.consumer, args=[callback])
        consumer_thread.start()

    def consumer(self, callback):
        while True:
            if self.qsize() > 0 or self._run:
                if callback:
                    callback(self.get())
                else:
                    self.get()

    def add(self, value):
        if self._run and value:
            self.put(value)
            
    def close(self):
        self._run = False

class X7MessageStream(object):
    def __init__(self) -> None:
        self.channels = {}

    def subscribe(self, key, maxsize = 1000, callback = None):
        if key is None:
            raise ValueError("Key is None")

        if key in self.channels.keys():
            logger.warning(f"Subscribe key {key} already exists.")

        # 每次订阅单独维护队列  
        self.channels[key] = MessageChannel(maxsize)
        self.channels[key].subscribe(callback)

    def remove(self, key):
        if self.channels[key]:
            self.channels[key].close()
            self.channels[key].remove(key)

    def _cache(self, packet):
        #将x7消息缓存到文件中
        return


    



from cmath import log
from ipaddress import ip_address
from socket import *
from time import strftime
from loguru import logger

class UDPSender(object):
    def __init__(self, ip = "255.255.255.255", port = 54366) :
        self._address = (ip, port)
        self._socket = None
        self.socket()
    
    def socket(self):
        if self._socket is None:
            self._socket = socket(AF_INET, SOCK_DGRAM)
            self._socket.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
        
        return self._socket
        
    def send(self, message):
        try:
            self._socket.sendto(message, self._address)
        except Exception as e:
            logger.error("Exception on UDPSender send.")
            logger.error(e)
    
    def close(self):
        if self._socket :
            self._close()


class UDPReceiver(object):
    def __init__(self, parser=None, channel=None, port=54366):
        self._run = False
        self._port = port
        self._socket = None
        self._parser = parser
        self._channel = channel

    def close(self):
        self._run = False

    def accept(self):
        if self._port is None:
            logger.error(f"UDPReceiver未设置监听端口！")
            return

        if self._run:
            logger.warning(f"UDPReceiver正在监听端口{self._port}, 本次操作已忽略。")
            return

        self._run = True
        if self._socket is None:
            self._socket = socket(AF_INET, SOCK_DGRAM)  
            self._socket.bind(('', self._port)) 

        while self._run:
            # print(f"等待从端口{self._port}获取数据...")
            udp_data = self._socket.recvfrom(1024) 
            self._parser.parse(udp_data[0], udp_data[1])
            
        logger.info(f"关闭端口{self._port}的连接")
        self._close()  #关闭套接字
    
    def add_connect(self, devinfo):
        print(devinfo)
        if self._channel and devinfo:
            self._channel.put(devinfo['dev_id'])




from .x7base import X7UDP, X7TCP

class X7Box(X7TCP):
    def __init__(self, tcp_port=19128, udp_port=54366, auto_search=False, callback=None):
        super().__init__(port=tcp_port, callback=callback)
        self._udp = X7UDP(port=udp_port, tcp_port=tcp_port, auto_search=auto_search)
    
    def connect(self, devno):
        self._udp.connect(devno=devno)



import hashlib
from threading import Thread
from time import sleep
from typing import Literal, overload
from loguru import logger

from qldev.device import X7MessageStream
from qldev.network import TCPServer
from .x7parser import *
from .x7utils import *
from .devinfo import *

class X7TCP(TCPServer):
    def __init__(self, port=19128, callback=None):
        super().__init__(port, callback)

    # trigger仅在脑电采集时生效
    # type是trigger的标识-int类型
    def trigger(self, type=0, devno = None):
        # 获取设备
        client = self.get_client(devno)
        if client is None:
            if devno:
                print(f"X7base({devno}) not found")
            return

        else:
            client.trigger(type)

    # id是音频编号，取值范围为[0-9]，表示第1-10个音频位置
    # volumn是音量大小，取值范围[1-10]
    def wav_trigger(self, id, volumn, devno = None):
        # 获取设备
        client = self.get_client(devno)
        if client is None:
            if devno:
                logger.warning(f"device({devno}) not found")
            return
            
        else:
            client.wav_trigger(id, volumn)
       
    # id是音频编号，取值范围为[0-9]，表示第1-10个音频位置
    # fname是本地音频文件的存储路径  
    def update_wav(self, id, fname, devno = None, buffer=3072):    
        # 获取设备
        client = self.get_client(devno)
        if client is None:
            if devno:
                print(f"X7base({devno}) not found")
            return

        with open(fname, 'rb') as f:
            flen = f.seek(0, 2)
            logger.debug(f"file{fname} len is {flen}")
            client.wav_update_ready(id, flen)
            sleep(0.1)

            f.seek(0)
            offset = 0
            buf = f.read(buffer)
            md5 = hashlib.md5()
            while buf:
                blen = len(buf)
                logger.debug(f"update(id:{id}, offset:{offset}/{flen}, llen:{blen})")
                client.wav_update(id, offset, blen, buf)
                md5.update(buf)
                # 指令发送间隔，避免数据处理异常
                sleep(0.01)
                offset += blen
                buf = f.read(buffer)
            logger.debug(f"file{fname} offset is {offset} md5 is {md5.hexdigest()}")

            client.wav_update_stop(id, md5.hexdigest())
            logger.debug(f'file{fname} update finished.')    

    def subscribe(self, key=None, devno=None, callback=None, maxsize=1000):
        # 获取设备
        client = self.get_client(devno)
        if client is None:
            if devno:
                print(f"X7base({devno}) not found")
            return
        
        if key is None:
            key = f"x7_{client.devid}"
        client.subscribe(f"x7_{key}", callback=callback, maxsize=maxsize) 

        
    # @overload
    def client_handler(self, socket, addr):
        logger.info(f"new client socket {addr[0]}:{addr[1]} connected.")
        ClientHandler(socket, MessageParser(), callback=self._client_changed)        
    
    def _client_changed(self, devno, type: Literal["add", "remove"], client = None):
        logger.info(f'{type} client {devno}')
        if type == 'add' and client:
            self.clients[devno] = client
            global_cache.connected_add(devno)
        else:
            self.clients.pop(devno)
            global_cache.connected_remove(devno)
        logger.info(f"Now has {len(self.clients)} clients.")

class ClientHandler(X7MessageStream):
    def __init__(self, socket, addr = None, parser = None, callback=None) :
        super().__init__()
        self._socket = socket
        self._address = addr
        self._parser = parser
        self._callback = callback
        self._run = True
        self.devid = None
        self.devtype = None
        self.accept()

    def accept(self):        
        # self.ask()
        # recv = Thread(target=client_listen, args=(self._socket, self.ask()))
        recv = Thread(target=self._consumer)
        recv.start()

        # 获取设备信息
        self.get_device_info()

    
    # 处理数据接收
    def _consumer(self):
        if self._parser is None:
            self._parser = MessageParser()

        while self._run:
            try:
                # 接收对方发送过来的数据
                recv_data = self._socket.recv(1024)  # 接收1024个字节
                if recv_data:
                    self._parser.parse(recv_data, self.update_info)
                else:
                    break
            except Exception as e:
                logger.error("Exception on ClientHandler._consumer()")
                logger.error(e)
                break

        # print(f"客户端{}:{}已关闭")
        self._socket.close()
        logger.info(f"device({self.devid}) disconnect.")
        if self._callback and self.devid:
            self._callback(self.devid, 'remove')


    def close(self):
        self._run = False

    def get_device_info(self):
        logger.info("ask the device info...")
        self._send(X7Command.get_device_info())

    def update_info(self, cmd, info):

        if cmd == CommandEnum.GET_DEVICE_INFO:
            self.devid = info['dev_id'] if info and 'dev_id' in info.keys() else None
            self.devtype = info['dev_type'] if info and 'dev_type' in info.keys() else None   
            if self._callback and self.devid:
                self._callback(self.devid, 'add', client = self)

        if cmd == CommandEnum.SIGNALS:
            if len(self.channels) > 0:
                for idx, item in enumerate(self.channels.keys()):
                    self.channels[item].put(info)

    def trigger(self, type=0):
        self._send(X7Command.trigger(type))

    def wav_update_ready(self, id, len):
        self._send(X7Command.wav_update_ready(id, len))

    def wav_update(self, id, offset, len, data):
        self._send(X7Command.wav_update(id, offset, len, data))

    def wav_update_stop(self, id, checksum):
        self._send(X7Command.wav_update_stop(id, checksum))

    def wav_trigger(self, id, volumn):
        self._send(X7Command.wav_trigger(id, volumn))

    def _send(self, message):
        try:
            self._socket.send(message)
        except Exception as e:
            logger.error("Exception on message send.")
            logger.error(e)


            



from struct import pack
from loguru import logger


from .x7utils import CommandEnum
from qldev.utils import crc16, get_ip


class MessageParser(object):
    def __init__(self) -> None:
        pass

    def parse(self, packet, callback = None):
        # print(f"MessageParser.parse. packet is {packet}")
        plen = len(packet)
        if plen < 10:
            logger.warning(f"MessageParser.parse. Not support message len: {len}")
            return
        
        cmd = int.from_bytes(packet[12:14], 'little')
        
        result = self.parse_by_cmd(cmd, packet[14:-2])
        if callback:
            callback(cmd, result)

    def parse_by_cmd(self, cmd, data):
        # 结果校验
        result = (int)(data[8])
        if result != 0:
            logger.warning(f"cmd {hex(cmd)} result is {result}")
            return None

        if cmd == CommandEnum.GET_DEVICE_INFO:
            return self.get_device_info(data)
            
        if cmd == CommandEnum.SIGNALS:
            return self.signals(data)
            
        if cmd == CommandEnum.TRIGGER:
            return self.trigger(data)
            
        if cmd == CommandEnum.WAV_UPDATE_READY:
            return self.wav_update_ready(data)
            
        if cmd == CommandEnum.WAV_UPDATE:
            return self.wav_update(data)
            
        if cmd == CommandEnum.WAV_UPDATE_STOP:
            return self.wav_update_stop(data)
            
        if cmd == CommandEnum.WAV_TRIGGER:
            return self.wav_trigger(data)

        return self.not_support(cmd, data)
        

    def get_device_info(self, data):
        info = {}
        info["dev_id"] = data[9:17].hex()
        info["dev_type"] = int.from_bytes(data[17:21], 'little')
        info["sw_version"] = int.from_bytes(data[21:25], 'little')
        info["hw_version"] = int.from_bytes(data[25:29], 'little')
        logger.debug(f"get_device_info->{info}")
        return info

    def signals(self, data):
        packet = {}
        packet["packet_id"] = int.from_bytes(data[9:17], 'little')
        packet["channels"] = data[17:49]
        packet["origin_rate"] = int.from_bytes(data[49:53], 'little')
        packet["sample_rate"] = int.from_bytes(data[53:57], 'little')
        packet["sample_length"] = int.from_bytes(data[57:61], 'little')
        packet["resolution"] = int.from_bytes(data[61:63], 'little')
        packet["filter"] = int(data[63])
        packet["date_len"] = int.from_bytes(data[64:68], 'little')
        packet["data"] = data[68:]

        return packet
        
    def trigger(self, data):
        logger.debug("trigger success.")
        return
        
    def wav_update_ready(self, data):
        logger.debug("wav_update_ready success.")
        return
        
    def wav_update(self, data):
        logger.debug("wav_update success.")
        return
        
    def wav_update_stop(self, data):
        logger.debug("wav_update_stop success.")
        return
        
    def wav_trigger(self, data):
        logger.debug("wav_trigger success.")
        return
        
    def not_support(self, cmd, data):
        logger.warning(f"cmd({cmd})not support.")
        return

        
class X7Parser(object):
    UDP_START = "SHQuanLan"
    NOTIFY_CMD = 0x10
    INFO_CMD = 0x09

    def __init__(self) -> None:
        self._base = None
        self._message = None

    def parse(self, packet):
        plen = len(packet)
        if plen < 10:
            return

        start = packet[:9].decode("utf-8")
        # quanlan udp message
        if start != X7Parser.UDP_START:
            logger.warning(f"unexcept message header.")
            return

        if not check_crc(packet):
            print(f"CRC verification failed!")
            return

        self._message = None
        cmd = int.from_bytes(packet[10:11], 'little')

        if cmd == 0x10:
            logger.debug(parse_net_param(packet[12:-2]))

        if cmd == 0x09:
            logger.debug(parse_net_param(packet[12:-2]))

    #devno是hex字符串
    @staticmethod
    def x7_notify(devno, devtype="f02a"):
        message = bytearray(28)
        message[:10] = X7Parser.UDP_START.encode('utf-8')
        cmd = 0x10
        dev_type = 0x38
        port = 54336
        ip = get_ip().split(".")
        
        message[10:12] = cmd.to_bytes(2, 'little')
        dev_id = int(devno, 16)
        message[12:20] = dev_id.to_bytes(8, 'little')
        # message[12:14] = bytes.fromhex(devtype)
        # #d57ee4dc99ce0940        
        # message[14:22] = bytes.fromhex(devno)
        message[22] = (int)(ip[0])
        message[23] = (int)(ip[1])
        message[24] = (int)(ip[2])
        message[25] = (int)(ip[3])
        message[26:28] = port.to_bytes(2, 'little')
        checksum = crc16(message)

        return message + checksum.to_bytes(2, 'little')


def check_crc(data):
    return int.from_bytes(data[-2:], 'little')  == crc16(data[:-2])

def get_data(cmd, packet):
    if cmd == 0x10:
        return packet[12:27]

    return packet[12:] 


def parse_net_param(data):
    dev = DeviceInfo()
    dev.unique_id = data[:10].hex()
    dev.dev_type = data[:2].hex()
    dev.dev_id = data[2:10].hex()
    dev.ip = f"{data[10]}.{data[11]}.{data[12]}.{data[13]}"
    dev.port = int.from_bytes(data[-2:], 'little')

    return dev


import json
class DeviceInfo():
    def __init__(self):
        self.unique_id = None
        self.dev_type = None
        self.dev_id = None
        self.ip = None
        self.port = None

    def __iter__(self):
        yield from {
            "unique_id": self.unique_id,
            "dev_type": self.dev_type,
            "dev_id": self.dev_id,
            "ip": self.ip,
            "port": self.port
        }.items()

    def __str__(self):
        return json.dumps(dict(self), ensure_ascii=False)

    def __repr__(self):
        return self.__str__()


from threading import Thread
from time import sleep
from qldev.device import UDPParser, MessageChannel, UDPMessage
from qldev.network import UDPReceiver, UDPSender
from .x7parser import *
from .devinfo import *

class X7UDP(object):
    def __init__(self, port = 54366, tcp_port=19128, auto_search = False, broadcast_interval = 3) :
        self._port = port
        self._sender = None
        self._receiver = None
        self._tcp_port = tcp_port
        self._list = []
        self._broadcast_interval = broadcast_interval
        self._create_server()
        if auto_search :
            self._create_client()

    # udp服务端
    def _create_server(self):
        if self._sender:
            return
        
        self._sender = UDPSender(port=self._port)
        # connect消息发送
        notify_thread = Thread(target=self._connect)
        notify_thread.start()

    # udp客户端
    def _create_client(self):
        if self._receiver:
            return
        
        channel = MessageChannel(maxsize=100)
        channel.subscribe(self.connect)
        parser = UDPParser(channel)
        
        self._receiver = UDPReceiver(parser=parser, port=self._port)
        # 消息监听
        client_thread = Thread(target=self._receiver.accept)
        client_thread.start()

    # 通过广播通知设备连接服务端
    def connect(self, devno):
        global_cache.broadcast_add(devno)

    
    # 发送广播
    def _connect(self):
        while True:
            list = global_cache.get_broadcast_list()
            for idx, item in enumerate(list):
                logger.info(f"search for device({item}).")
                message = UDPMessage.get_connect(item, self._tcp_port)            
                self._sender.send(message=message)

            # 广播间歇
            if self._broadcast_interval >  0:
                sleep(self._broadcast_interval)
    
    # 开启客户端
    def open_client(self):
        self._create_client()

    # 清空广播列表
    def clear_broadcast(self, type: Literal['all', 'tcp', 'udp'] = 'udp'):
        global_cache.clear(type)

    def close(self):
        if self._sender:
            self._sender.close()

        if self._receiver:
            self._receiver.close()



from .message import UDPMessage
from qldev.utils import crc16


class MessageParser(object):
    def __init__(self) -> None:
        pass

class UDPParser(MessageParser):
    def __init__(self, channel = None) -> None:
        super().__init__()
        self._channel = channel

    def parse(self, packet, address):
        # print(f"receive udp message from {address[0]}:{address[1]}")
        plen = len(packet)
        # print(f"parse {packet} len {plen}")
        if plen < 10:
            return

        start = packet[:9].decode("utf-8")
        # quanlan udp message
        if start != UDPMessage.MessageStart:
            return

        if not check_crc(packet):
            print(f"数据CRC校验失败，丢弃！")
            return
        
        # message command
        cmd = int.from_bytes(packet[10:12], 'little')

        return self._parse(cmd, packet[12:])

    def _parse(self, cmd, data):
        # 只解析0x09
        if cmd == UDPMessage.DeviceInfoCommand:
            return self.parseDeviceInfo(data)
        else:
            # print(f'Ignore message which command is {hex(cmd)}')
            return None
        

    def parseDeviceInfo(self, data):
        # print(f"parseDeviceInfo {data}")
        devinfo = {}
        try:
            devinfo["dev_type"] = hex(int.from_bytes(data[:2], 'little'))
            devinfo["dev_id"] = data[2:10].hex()
            devinfo["version_code"] = int.from_bytes(data[42:46], 'little')
            devinfo["version_name"] = str(data[10:42],'utf-8').split('\x00')[0]
            if self._channel:
                self._channel.add(str(devinfo["dev_id"]))

        except Exception as e:
            print(f"parseDeviceInfo异常：{e}")

        # print(f"dev info is {devinfo}")
        return devinfo


def check_crc(data):
    return int.from_bytes(data[-2:], 'little')  == crc16(data[:-2])





from qldev.utils import crc16

class CommandEnum():
    GET_DEVICE_INFO = 0x10
    SIGNALS = 0x455
    TRIGGER = 0x501
    WAV_UPDATE_READY = 0x502
    WAV_UPDATE = 0x503
    WAV_UPDATE_STOP = 0x504
    WAV_TRIGGER = 0x505

    
class CommandHeader():
    start = 0xa55a
    def __init__(self) -> None:
        pass
    
    @staticmethod
    def get(packet_type=2, dev_type=0, dev_id = 0):
        header = bytearray(12)
        header[:2] = CommandHeader.start.to_bytes(2, 'little')
        header[2] = packet_type
        header[3] = 42
        header[4:8] = dev_id.to_bytes(4, 'little')
        len = 16
        header[8:12] = len.to_bytes(4, 'little')
        # message[12:14] = CommandEnum.GET_DEVICE_INFO.to_bytes(2, 'little')
    

        return  header
    
class DeviceCommand(object):
    def __init__(self) -> None:
        pass

    @staticmethod
    def get_device_info():
        return DeviceCommand.wrap(CommandEnum.GET_DEVICE_INFO)

    @staticmethod
    def header():
        return b'0x5aa5'

    @staticmethod
    def wrap(cmd, data=None):
        header = CommandHeader.get()
        if data:
            dlen = len(data) + 16
            header[8:12] = dlen.to_bytes(4, 'little')
            message = header + cmd.to_bytes(2, 'little') + data 
        else: 
            message = header + cmd.to_bytes(2, 'little')
        
        checksum = crc16(message)
        return message + checksum.to_bytes(2, 'little')

def to_bytes(value, len=4, byteorder = 'little'):
    re = bytearray(len)
    for i in range(len):
        if byteorder == 'little':
            re[i] = (value >> (8 * i)) & 0xFF
        else:
            idx = len - i - 1
            re[idx] = (value >> (8 * i)) & 0xFF
    return re

class X7Command(DeviceCommand):
    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def trigger(type = 0):
        data = type.to_bytes(4, 'little')
        return DeviceCommand.wrap(CommandEnum.TRIGGER, data)

    @staticmethod
    def wav_update_ready(id, len):
        data = bytearray(8)
        data[:4] = id.to_bytes(4, 'little')
        data[4:] = len.to_bytes(4, 'little')
        return DeviceCommand.wrap(CommandEnum.WAV_UPDATE_READY, data)

    @staticmethod
    def wav_update(id, offset, len, content):
        data = bytearray(len + 12)
        data[:4] = id.to_bytes(4, 'little')
        data[4:8] = offset.to_bytes(4, 'little')
        data[8:12] = len.to_bytes(4, 'little')
        data[12:] = content
        return DeviceCommand.wrap(CommandEnum.WAV_UPDATE, data)

    @staticmethod
    def wav_update_stop(id, md5sum):
        data = id.to_bytes(4, 'little')
        data += md5sum.encode("utf-8") + b'0x0' # 结尾符号解析
        return DeviceCommand.wrap(CommandEnum.WAV_UPDATE_STOP, data)

    @staticmethod
    def wav_trigger(id = 0, volumn = 2):
        data = bytearray(8)
        data[:4] = id.to_bytes(4, 'little')
        data[4:] = volumn.to_bytes(4, 'little')
        return DeviceCommand.wrap(CommandEnum.WAV_TRIGGER, data)



from typing import Literal


class X7Cache(object):
    def __init__(self) :
        self._dev_broadcast_list = set([])
        self._dev_connected_list = set([])

    def broadcast_add(self, devno):
        self._dev_broadcast_list.add(devno)

    def broadcast_remove(self, devno):
        self._dev_broadcast_list.remove(devno)

    def connected_add(self, devno):
        self._dev_connected_list.add(devno)

    def connected_remove(self, devno):
        self._dev_connected_list.remove(devno)

    def get_broadcast_list(self):
        # 已连接的设备不再广播
        return self._dev_broadcast_list.difference(self._dev_connected_list)

    def clear(self, type : Literal['all', 'tcp', 'udp'] = 'udp'):

        if type == 'udp' or type == 'all' :
            self._dev_broadcast_list.clear()

        if type == 'tcp' or type == 'all' :
            self._dev_connected_list.clear()

# 全局对象
global_cache = X7Cache()


    


